import enum
import functools
from typing import Union, Tuple

import stringcase
from django.apps import apps
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from graphql_jwt import decorators
from graphql_jwt.exceptions import PermissionDenied

from django_koldar_utils.django import django_helpers


class PermissionClass(enum.Enum):
    """
    An enumeration that list all the possible actions a permission can use
    """
    CREATE = "add"
    READ = "view"
    UPDATE = "change"
    DELETE = "delete"


def get_permission_codename(action: PermissionClass, model: type) -> str:
    """
    Fetch the code name of a permission

    :param action: the action a standard permission should allow
    :param model: model involved
    :return: permission code name that can be used in permission_required
    """
    return f"{action.value}_{stringcase.lowercase(model.__name__)}"


def get_permission_create_name(model: type) -> str:
    """
    Fetch the code name of a create permission

    :param model: model involved
    :return: permission code name that can be used in permission_required
    """
    return get_permission_codename(PermissionClass.CREATE, model)


def get_permission_read_name(model: type) -> str:
    """
    Fetch the code name of a read permission

    :param model: model involved
    :return: permission code name that can be used in permission_required
    """
    return get_permission_codename(PermissionClass.READ, model)


def get_permission_update_name(model: type) -> str:
    """
    Fetch the code name of an update permission

    :param model: model involved
    :return: permission code name that can be used in permission_required
    """
    return get_permission_codename(PermissionClass.UPDATE, model)


def get_permission_delete_name(model: type) -> str:
    """
    Fetch the code name of a delete permission

    :param model: model involved
    :return: permission code name that can be used in permission_required
    """
    return get_permission_codename(PermissionClass.DELETE, model)


def create_permission_to_save(codename: str, model: type, name: str = None) -> Permission:
    """
    Generate  anew instance of Permission. This is very handy if you wnat to programmatically create permissions.
    Just save the instance generated by this method.

    :param name: name of hte permission. If left unspecifid, it is derived from codename
    :param codename: codename of the permission
    :param in_migration: true if we are runnign this code in a migration. false otherwise
    :param model: model that owns the permission
    :return: a new instance of permission
    """

    content_type = ContentType.objects.get_for_model(model)
    if django_helpers.are_we_in_migration():
        permissions_class = apps.get_model('auth', 'Permission')
    else:
        permissions_class = Permission
    if name is None:
        name = stringcase.sentencecase(codename)
    permission = permissions_class(
        codename=codename,
        name=name,
        content_type=content_type,
    )
    return permission


def create_saved_crud_permissions(model: type, concept: str = None) -> Tuple[Permission, Permission, Permission, Permission]:
    """
    Create the basic 4 crud permission given a a string.
    We will store such 4 permissions inside the databae immediately, so you don't need to asve them.

    The permissions will be saved as "can_create_x", "can_view_x", "can_update_x", "can_delete_x" where
    x is a representation of "model"

    :param model: a django model type models.Model
    :param concept: name fo the permission concept. If left missing, it is model
    :return: 4 permissions CRUD
    """

    if concept is None:
        concept = model.__name__
    create = create_permission_to_save(
        codename=f"can_create_{stringcase.snakecase(concept)}",
        name=f"Can create {stringcase.sentencecase(concept)}",
        model=model
    ).save()
    read = create_permission_to_save(
        codename=f"can_view_{stringcase.snakecase(concept)}",
        name=f"Can view {stringcase.sentencecase(concept)}",
        model=model
    ).save()
    update = create_permission_to_save(
        codename=f"can_update_{stringcase.snakecase(concept)}",
        name=f"Can update {stringcase.sentencecase(concept)}",
        model=model
    ).save()
    delete = create_permission_to_save(
        codename=f"can_delete_{stringcase.snakecase(concept)}",
        name=f"Can delete {stringcase.sentencecase(concept)}",
        model=model
    ).save()
    return create, read, update, delete



def ensure_user_has_permissions(perm):
    """
    A copy of permission_required where the error generates the missing permissions
    """

    def decorator(f):
        @functools.wraps(f)
        @decorators.context(f)
        def wrapper(context, *args, **kwargs):
            if isinstance(perm, str):
                perms = (perm,)
            else:
                perms = perm
            missing_permissions = list(filter(lambda p: not context.user.has_perm(p), perms))
            if len(missing_permissions) > 0:
                raise PermissionDenied(
                    f"The user {context.user} do not have the required permissions {', '.join(missing_permissions)} for accessing the function.")
            return f(*args, **kwargs)
        return wrapper
    return decorator
